

class Territory:
    """ String """
    name = None

    def __init__(self, name):
        self.name = name

    def same_territory(self, other):
        return self.name == other.name


class CoastTerritory(Territory):
    """ LandTerritory """
    parent = None

    def __init__(self, name, parent):
        super().__init__(name)
        assert isinstance(parent, LandTerritory)
        self.parent = parent

    def same_territory(self, other):
        relevant_territories = (
            { self.parent.name } |
            { coast.name for coast in self.parent.coasts }
        )
        return other.name in relevant_territories

    def __eq__(self, other):
        if not isinstance(other, CoastTerritory):
            return False

        if self.name != other.name:
            return False

        if self.parent.name != other.parent.name:
            return False

        self_parent_coasts  = { coast.name for coast in self.parent.coasts }
        other_parent_coasts = { coast.name for coast in other.parent.coasts }
        if self_parent_coasts != other_parent_coasts:
            return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class LandTerritory(Territory):
    """ CoastTerritory[] """
    coasts = None

    def __init__(self, name, coast_names):
        super().__init__(name)
        self.coasts = []
        for coast_name in coast_names:
            self.coasts.append(CoastTerritory(coast_name, self))

    def same_territory(self, other):
        relevant_territories = (
            { self.name } |
            { coast.name for coast in self.coasts }
        )
        return other.name in relevant_territories

    def __eq__(self, other):
        if not isinstance(other, LandTerritory):
            return False

        if self.name != other.name:
            return False

        self_coasts  = { coast.name for coast in self.coasts }
        other_coasts = { coast.name for coast in other.coasts }
        if self_coasts != other_coasts:
            return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class SeaTerritory(Territory):
    def __init__(self, name):
        super().__init__(name)

    def __eq__(self, other):
        if not isinstance(other, SeaTerritory):
            return False
        return self.name == other.name

    def __ne__(self, other):
        return not self.__eq__(other)



from pydip.map.territory import LandTerritory, SeaTerritory, CoastTerritory


class Map:
    """ String -> Territory """
    name_map = None

    """ String -> String{} (Adjacency List) """
    adjacency = None

    """
    territory_descriptors: list of structs defining new territories. Of the form:
        {
            'name': String,
            'coasts': [ { 'name': String } ] (Optional)
        }

        If 'coasts' is provided, the territory is Land, with Coasts as children. Otherwise it is Sea.

    adjacencies: list of tuples of names representing adjacent territories. Must be legal:
        * Sea adjacent only to Sea or Coast
        * Coast adjacent only to Sea or Coast
        * Land adjacent only to Land
        * All names must be provided in territory_descriptors
        * No duplicate adjacencies!
    """
    def __init__(self, territory_descriptors, adjacencies):
        self.name_map = dict()
        self.adjacency = dict()

        self._setup_name_map(territory_descriptors)
        self._setup_adjacencies(adjacencies)

    def relevant_name_for_territory(self, territory_name):
        territory = self.name_map[territory_name]
        if isinstance(territory, CoastTerritory):
            territory_name = territory.parent.name
        return territory_name

    def _setup_name_map(self, territory_descriptors):
        for descriptor in territory_descriptors:
            assert 'name' in descriptor
            name = descriptor['name']
            if 'coasts' in descriptor:
                land = LandTerritory(name, descriptor['coasts'])
                self._add_territory(land)
                for coast in land.coasts:
                    self._add_territory(coast)
            else:
                self._add_territory(SeaTerritory(name))

    def _add_territory(self, territory):
        self.name_map[territory.name] = territory
        self.adjacency[territory.name] = set()

    def _setup_adjacencies(self, adjacencies):
        for name_a, name_b in adjacencies:
            assert name_a in self.name_map
            assert name_b in self.name_map

            territory_a = self.name_map[name_a]
            territory_b = self.name_map[name_b]

            assert ~(isinstance(territory_a, LandTerritory) ^ isinstance(territory_b, LandTerritory))
            assert name_b not in self.adjacency[name_a]
            assert name_a not in self.adjacency[name_b]

            self.adjacency[name_a].add(name_b)
            self.adjacency[name_b].add(name_a)


class SupplyCenterMap:
    """ Map """
    game_map = None

    """ String{}, subset of map.name_map.keys() """
    supply_centers = None

    def __init__(self, game_map, supply_centers):
        assert supply_centers < game_map.name_map.keys()

        self.supply_centers = set(supply_centers)
        self.game_map = game_map


class OwnershipMap:
    """ SupplyCenterMap """
    supply_map = None

    """ String -> String{}, mapping Player names to which territories they control """
    owned_territories = None

    """ String -> String{}, mapping Player names to which territories are their home territories """
    home_territories = None

    def __init__(self, supply_map, owned_territories, home_territories):
        players = owned_territories.keys()
        assert all(owned_territories[player] - supply_map.supply_centers == set() for player in players)
        assert all(home_territories[player] - supply_map.supply_centers == set() for player in players)

        self.supply_map = supply_map
        self.owned_territories = owned_territories
        self.home_territories = home_territories

    def territory_is_owned(self, player, territory_name):
        relevant_name = self.supply_map.game_map.relevant_name_for_territory(territory_name)
        return relevant_name in self.owned_territories[player]

    def territory_is_home(self, player, territory_name):
        relevant_name = self.supply_map.game_map.relevant_name_for_territory(territory_name)
        return relevant_name in self.home_territories[player]



from enum import Enum


class UnitTypes(Enum):
    TROOP = 0
    FLEET = 1


class Unit:
    """ UnitType """
    unit_type = None

    """ String -- Name of occupied territory """
    position = None

    def __init__(self, unit_type, position):
        self.unit_type = unit_type
        self.position = position

    def __eq__(self, other):
        return self.unit_type == other.unit_type and self.position == other.position

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.unit_type, self.position))

    def __repr__(self):
        return '[{} -- {}]'.format(self.unit_type, self.position)



from pydip.player.command.retreat_command import RetreatCommand, RetreatMoveCommand
from pydip.player.unit import Unit


def resolve_retreats(retreat_map, commands):
    """
    Returns resulting board state by considering interactions of provided list
    of retreat commands.

    retreat_map is intended to be the output from turn.resolve.resolve_turn.
    That is, a mapping of players to mappings of units to either None (in the
    case that a retreat is not expected for that unit), or a set of territory
    names (which hold the valid retreat targets).

    Returns a mapping of players to lists of units, representing which units
    in which locations those players will have after resolving retreats. Will
    be equivalent to the entries in retreat_map, minus any disbanded units --
    and, of course, without any retreat requirements.
    """
    assert all(isinstance(command, RetreatCommand) for command in commands)
    retreaters = { (command.player.name, command.unit) for command in commands }
    valid_retreaters = set()
    for player in retreat_map.keys():
        valid_retreaters |= {
            (player, unit)
            for unit in retreat_map[player].keys()
            if retreat_map[player][unit] is not None
        }
    assert retreaters == valid_retreaters

    result_map = { player : set() for player in retreat_map }
    for player in retreat_map:
        for unit in retreat_map[player]:
            if retreat_map[player][unit] is None:
                result_map[player].add(unit)

    commands = list(filter(lambda c: isinstance(c, RetreatMoveCommand), commands))
    for command in commands:
        other_commands = list(filter(lambda c: c != command, commands))
        if all(command.destination != other_command.destination for other_command in other_commands):
            result_map[command.player.name].add(Unit(command.unit.unit_type, command.destination))

    return result_map



from collections import defaultdict

from pydip.player.command.command import MoveCommand, ConvoyMoveCommand, ConvoyTransportCommand, SupportCommand


class CommandMap:

    """ territory -> MoveCommand[], representing commands with destination at key territory """
    _attacker_map        = None
    """ territory -> ConvoyMoveCommand[], representing commands with destination at key territory """
    _convoy_attacker_map = None
    """ (source, dest) -> ConvoyTransportCommand[], representing transport commands from source to dest territories """
    _transport_map       = None
    """ (source, dest) -> SupportCommand[], representing supports of units attacking from source to dest territories """
    _support_map         = None
    """ territory -> Command, representing command for unit originating at key territory """
    _home_map            = None

    _game_map = None

    """ Builds CommandMap from provided Command[] """
    def __init__(self, game_map, commands):
        self._game_map = game_map
        self._attacker_map        = defaultdict(list)
        self._convoy_attacker_map = defaultdict(list)
        self._transport_map       = defaultdict(list)
        self._support_map         = defaultdict(list)
        self._home_map            = dict()

        for command in commands:
            home_name = self._game_map.relevant_name_for_territory(command.unit.position)
            self._home_map[home_name] = command
            if isinstance(command, MoveCommand):
                destination_name = self._game_map.relevant_name_for_territory(command.destination)
                self._attacker_map[destination_name].append(command)
            elif isinstance(command, ConvoyMoveCommand):
                destination_name = self._game_map.relevant_name_for_territory(command.destination)
                self._convoy_attacker_map[destination_name].append(command)
            elif isinstance(command, ConvoyTransportCommand):
                source = self._game_map.relevant_name_for_territory(command.transported_unit.position)
                dest   = self._game_map.relevant_name_for_territory(command.destination)
                self._transport_map[(source, dest)].append(command)
            elif isinstance(command, SupportCommand):
                source = self._game_map.relevant_name_for_territory(command.supported_unit.position)
                dest   = self._game_map.relevant_name_for_territory(command.destination)
                self._support_map[(source, dest)].append(command)

    def get_attackers(self, territory_name):
        territory_name = self._game_map.relevant_name_for_territory(territory_name)
        return self._attacker_map[territory_name]

    def get_convoy_attackers(self, territory_name):
        territory_name = self._game_map.relevant_name_for_territory(territory_name)
        return self._convoy_attacker_map[territory_name]

    def get_convoy_transports(self, source_name, destination_name):
        source_name = self._game_map.relevant_name_for_territory(source_name)
        destination_name = self._game_map.relevant_name_for_territory(destination_name)
        return self._transport_map[(source_name, destination_name)]

    def get_supports(self, source_name, destination_name):
        source_name = self._game_map.relevant_name_for_territory(source_name)
        destination_name = self._game_map.relevant_name_for_territory(destination_name)
        return self._support_map[(source_name, destination_name)]

    def get_home_command(self, territory_name):
        territory_name = self._game_map.relevant_name_for_territory(territory_name)
        return self._home_map.get(territory_name, None)



from pydip.player.unit import Unit
from pydip.player.helpers import unit_type_can_enter
from pydip.map.territory import CoastTerritory


class Player:
    """ String -- name identifier for player """
    name = None

    """ Map -- reference to game board """
    game_map = None

    """ Unit[] -- units owned by this player """
    units = None

    """ String{} -- names of starting territories for player """
    starting_territories = None

    """
     * name: String -- identifier for player
     * map:  Map    -- reference to game board
     * starting_configuration: Defines how player should begin configuration. List of structs of form:
        { 'territory_name': String, 'unit_type': UnitType? (None indicates no unit) }
        - Duplicate territory definitions are not allowed!
        - Coasts should be provided as starting territories for fleets, but the parent will be labeled as the actual
          starting territory.
    """
    def __init__(self, name, game_map, starting_configuration):
        self.name = name
        self.game_map = game_map
        self.starting_territories = set()
        self.units = []

        for config in starting_configuration:
            name = config['territory_name']
            unit_type = config['unit_type']
            assert name in self.game_map.name_map.keys()

            territory = self.game_map.name_map[name]
            starting_territory = territory.parent if isinstance(territory, CoastTerritory) else territory
            assert starting_territory.name in self.game_map.name_map.keys()
            assert starting_territory.name not in self.starting_territories
            self.starting_territories.add(starting_territory.name)

            if unit_type is None:
                continue
            assert unit_type_can_enter(unit_type, territory)

            self.units.append(Unit(unit_type, name))

        assert len(self.starting_territories) == len(starting_configuration)



from copy import deepcopy

from pydip.map.map import Map, SupplyCenterMap, OwnershipMap
from pydip.player.unit import Unit
from pydip.player.unit import UnitTypes


_VANILLA_DIP_MAP_CACHE = None
_VANILLA_DIP_SUPPLY_CENTER_MAP_CACHE = None
_VANILLA_DIP_HOME_TERRITORY_CACHE = None
_VANILLA_DIP_STARTING_OWNERSHIP_MAP_CACHE = None
_VANILLA_DIP_STARTING_PLAYER_UNITS_CACHE = None


def generate_map():
    global _VANILLA_DIP_MAP_CACHE
    if not _VANILLA_DIP_MAP_CACHE:
        territory_descriptors = [
            # Sea Territories
            {'name': 'Adriatic Sea'},
            {'name': 'Aegean Sea'},
            {'name': 'Baltic Sea'},
            {'name': 'Barents Sea'},
            {'name': 'Black Sea'},
            {'name': 'Eastern Mediterranean Sea'},
            {'name': 'English Channel'},
            {'name': 'Gulf of Bothnia'},
            {'name': 'Gulf of Lyon'},
            {'name': 'Helgoland Bight'},
            {'name': 'Ionian Sea'},
            {'name': 'Irish Sea'},
            {'name': 'Mid-Atlantic Ocean'},
            {'name': 'North Atlantic Ocean'},
            {'name': 'North Sea'},
            {'name': 'Norwegian Sea'},
            {'name': 'Skagerrak'},
            {'name': 'Tyrrhenian Sea'},
            {'name': 'Western Mediterranean Sea'},

            # Land Territories
            {'name': 'Albania',        'coasts': ['Albania Coast']},
            {'name': 'Apulia',         'coasts': ['Apulia Coast']},
            {'name': 'Ankara',         'coasts': ['Ankara Coast']},
            {'name': 'Armenia',        'coasts': ['Armenia Coast']},
            {'name': 'Belgium',        'coasts': ['Belgium Coast']},
            {'name': 'Berlin',         'coasts': ['Berlin Coast']},
            {'name': 'Bohemia',        'coasts': []},
            {'name': 'Brest',          'coasts': ['Brest Coast']},
            {'name': 'Budapest',       'coasts': []},
            {'name': 'Bulgaria',       'coasts': ['Bulgaria North Coast', 'Bulgaria South Coast']},
            {'name': 'Burgundy',       'coasts': []},
            {'name': 'Clyde',          'coasts': ['Clyde Coast']},
            {'name': 'Constantinople', 'coasts': ['Constantinople Coast']},
            {'name': 'Denmark',        'coasts': ['Denmark Coast']},
            {'name': 'Edinburgh',      'coasts': ['Edinburgh Coast']},
            {'name': 'Finland',        'coasts': ['Finland Coast']},
            {'name': 'Gascony',        'coasts': ['Gascony Coast']},
            {'name': 'Galicia',        'coasts': []},
            {'name': 'Greece',         'coasts': ['Greece Coast']},
            {'name': 'Holland',        'coasts': ['Holland Coast']},
            {'name': 'Kiel',           'coasts': ['Kiel Coast']},
            {'name': 'London',         'coasts': ['London Coast']},
            {'name': 'Liverpool',      'coasts': ['Liverpool Coast']},
            {'name': 'Livonia',        'coasts': ['Livonia Coast']},
            {'name': 'Marseilles',     'coasts': ['Marseilles Coast']},
            {'name': 'Moscow',         'coasts': []},
            {'name': 'Munich',         'coasts': []},
            {'name': 'Naples',         'coasts': ['Naples Coast']},
            {'name': 'Norway',         'coasts': ['Norway Coast']},
            {'name': 'Paris',          'coasts': []},
            {'name': 'Picardy',        'coasts': ['Picardy Coast']},
            {'name': 'Piedmont',       'coasts': ['Piedmont Coast']},
            {'name': 'Portugal',       'coasts': ['Portugal Coast']},
            {'name': 'Prussia',        'coasts': ['Prussia Coast']},
            {'name': 'Rome',           'coasts': ['Rome Coast']},
            {'name': 'Ruhr',           'coasts': []},
            {'name': 'North Africa',   'coasts': ['North Africa Coast']},
            {'name': 'Rumania',        'coasts': ['Rumania Coast']},
            {'name': 'Serbia',         'coasts': []},
            {'name': 'Silesia',        'coasts': []},
            {'name': 'Sevastopol',     'coasts': ['Sevastopol Coast']},
            {'name': 'Smyrna',         'coasts': ['Smyrna Coast']},
            {'name': 'Spain',          'coasts': ['Spain North Coast', 'Spain South Coast']},
            {'name': 'St. Petersburg', 'coasts': ['St. Petersburg North Coast', 'St. Petersburg South Coast']},
            {'name': 'Sweden',         'coasts': ['Sweden Coast']},
            {'name': 'Syria',          'coasts': ['Syria Coast']},
            {'name': 'Trieste',        'coasts': ['Trieste Coast']},
            {'name': 'Tunis',          'coasts': ['Tunis Coast']},
            {'name': 'Tuscany',        'coasts': ['Tuscany Coast']},
            {'name': 'Tyrolia',        'coasts': []},
            {'name': 'Ukraine',        'coasts': []},
            {'name': 'Venice',         'coasts': ['Venice Coast']},
            {'name': 'Vienna',         'coasts': []},
            {'name': 'Warsaw',         'coasts': []},
            {'name': 'Wales',          'coasts': ['Wales Coast']},
            {'name': 'Yorkshire',      'coasts': ['Yorkshire Coast']},
        ]
        adjacencies = [
            # Sea Adjacencies
            ('North Atlantic Ocean',      'Norwegian Sea'),
            ('North Atlantic Ocean',      'Mid-Atlantic Ocean'),
            ('North Atlantic Ocean',      'Irish Sea'),
            ('North Atlantic Ocean',      'Clyde Coast'),
            ('North Atlantic Ocean',      'Liverpool Coast'),
            ('Norwegian Sea',             'North Sea'),
            ('Norwegian Sea',             'Barents Sea'),
            ('Norwegian Sea',             'Clyde Coast'),
            ('Norwegian Sea',             'Edinburgh Coast'),
            ('Norwegian Sea',             'Norway Coast'),
            ('Barents Sea',               'Norway Coast'),
            ('Barents Sea',               'St. Petersburg North Coast'),
            ('Mid-Atlantic Ocean',        'Irish Sea'),
            ('Mid-Atlantic Ocean',        'English Channel'),
            ('Mid-Atlantic Ocean',        'Western Mediterranean Sea'),
            ('Mid-Atlantic Ocean',        'North Africa Coast'),
            ('Mid-Atlantic Ocean',        'Spain North Coast'),
            ('Mid-Atlantic Ocean',        'Spain South Coast'),
            ('Mid-Atlantic Ocean',        'Portugal Coast'),
            ('Mid-Atlantic Ocean',        'Gascony Coast'),
            ('Mid-Atlantic Ocean',        'Brest Coast'),
            ('Irish Sea',                 'English Channel'),
            ('Irish Sea',                 'Liverpool Coast'),
            ('Irish Sea',                 'Wales Coast'),
            ('English Channel',           'Wales Coast'),
            ('English Channel',           'London Coast'),
            ('English Channel',           'North Sea'),
            ('English Channel',           'Belgium Coast'),
            ('English Channel',           'Picardy Coast'),
            ('English Channel',           'Brest Coast'),
            ('North Sea',                 'Skagerrak'),
            ('North Sea',                 'Helgoland Bight'),
            ('North Sea',                 'Norway Coast'),
            ('North Sea',                 'Holland Coast'),
            ('North Sea',                 'Belgium Coast'),
            ('North Sea',                 'London Coast'),
            ('North Sea',                 'Yorkshire Coast'),
            ('North Sea',                 'Edinburgh Coast'),
            ('North Sea',                 'Denmark Coast'),
            ('Skagerrak',                 'Norway Coast'),
            ('Skagerrak',                 'Sweden Coast'),
            ('Skagerrak',                 'Denmark Coast'),
            ('Helgoland Bight',           'Denmark Coast'),
            ('Helgoland Bight',           'Kiel Coast'),
            ('Helgoland Bight',           'Holland Coast'),
            ('Baltic Sea',                'Sweden Coast'),
            ('Baltic Sea',                'Gulf of Bothnia'),
            ('Baltic Sea',                'Livonia Coast'),
            ('Baltic Sea',                'Prussia Coast'),
            ('Baltic Sea',                'Berlin Coast'),
            ('Baltic Sea',                'Kiel Coast'),
            ('Baltic Sea',                'Denmark Coast'),
            ('Gulf of Bothnia',           'Finland Coast'),
            ('Gulf of Bothnia',           'St. Petersburg South Coast'),
            ('Gulf of Bothnia',           'Livonia Coast'),
            ('Gulf of Bothnia',           'Sweden Coast'),
            ('Western Mediterranean Sea', 'Spain South Coast'),
            ('Western Mediterranean Sea', 'Gulf of Lyon'),
            ('Western Mediterranean Sea', 'Tyrrhenian Sea'),
            ('Western Mediterranean Sea', 'Tunis Coast'),
            ('Western Mediterranean Sea', 'North Africa Coast'),
            ('Gulf of Lyon',              'Marseilles Coast'),
            ('Gulf of Lyon',              'Spain South Coast'),
            ('Gulf of Lyon',              'Piedmont Coast'),
            ('Gulf of Lyon',              'Tuscany Coast'),
            ('Gulf of Lyon',              'Tyrrhenian Sea'),
            ('Tyrrhenian Sea',            'Tuscany Coast'),
            ('Tyrrhenian Sea',            'Rome Coast'),
            ('Tyrrhenian Sea',            'Naples Coast'),
            ('Tyrrhenian Sea',            'Ionian Sea'),
            ('Tyrrhenian Sea',            'Tunis Coast'),
            ('Ionian Sea',                'Tunis Coast'),
            ('Ionian Sea',                'Naples Coast'),
            ('Ionian Sea',                'Apulia Coast'),
            ('Ionian Sea',                'Adriatic Sea'),
            ('Ionian Sea',                'Albania Coast'),
            ('Ionian Sea',                'Greece Coast'),
            ('Ionian Sea',                'Aegean Sea'),
            ('Ionian Sea',                'Eastern Mediterranean Sea'),
            ('Aegean Sea',                'Greece Coast'),
            ('Aegean Sea',                'Bulgaria South Coast'),
            ('Aegean Sea',                'Constantinople Coast'),
            ('Aegean Sea',                'Smyrna Coast'),
            ('Aegean Sea',                'Eastern Mediterranean Sea'),
            ('Eastern Mediterranean Sea', 'Smyrna Coast'),
            ('Eastern Mediterranean Sea', 'Syria Coast'),
            ('Adriatic Sea',              'Apulia Coast'),
            ('Adriatic Sea',              'Venice Coast'),
            ('Adriatic Sea',              'Trieste Coast'),
            ('Adriatic Sea',              'Albania Coast'),
            ('Black Sea',                 'Constantinople Coast'),
            ('Black Sea',                 'Bulgaria North Coast'),
            ('Black Sea',                 'Rumania Coast'),
            ('Black Sea',                 'Sevastopol Coast'),
            ('Black Sea',                 'Armenia Coast'),
            ('Black Sea',                 'Ankara Coast'),

            # Coast Adjacencies
            ('Clyde Coast',          'Edinburgh Coast'),
            ('Clyde Coast',          'Liverpool Coast'),
            ('Edinburgh Coast',      'Yorkshire Coast'),
            ('Liverpool Coast',      'Wales Coast'),
            ('Wales Coast',          'London Coast'),
            ('Yorkshire Coast',      'London Coast'),
            ('Norway Coast',         'Sweden Coast'),
            ('Norway Coast',         'St. Petersburg North Coast'),
            ('Sweden Coast',         'Denmark Coast'),
            ('Sweden Coast',         'Finland Coast'),
            ('Denmark Coast',        'Kiel Coast'),
            ('Finland Coast',        'St. Petersburg South Coast'),
            ('Livonia Coast',        'St. Petersburg South Coast'),
            ('Livonia Coast',        'Prussia Coast'),
            ('Prussia Coast',        'Berlin Coast'),
            ('Berlin Coast',         'Kiel Coast'),
            ('Kiel Coast',           'Holland Coast'),
            ('Holland Coast',        'Belgium Coast'),
            ('Belgium Coast',        'Picardy Coast'),
            ('Picardy Coast',        'Brest Coast'),
            ('Brest Coast',          'Gascony Coast'),
            ('Gascony Coast',        'Spain North Coast'),
            ('Portugal Coast',       'Spain North Coast'),
            ('Portugal Coast',       'Spain South Coast'),
            ('Marseilles Coast',     'Spain South Coast'),
            ('Marseilles Coast',     'Piedmont Coast'),
            ('Piedmont Coast',       'Tuscany Coast'),
            ('Tuscany Coast',        'Rome Coast'),
            ('Rome Coast',           'Naples Coast'),
            ('Naples Coast',         'Apulia Coast'),
            ('Apulia Coast',         'Venice Coast'),
            ('North Africa Coast',   'Tunis Coast'),
            ('Venice Coast',         'Trieste Coast'),
            ('Trieste Coast',        'Albania Coast'),
            ('Albania Coast',        'Greece Coast'),
            ('Greece Coast',         'Bulgaria South Coast'),
            ('Constantinople Coast', 'Bulgaria South Coast'),
            ('Constantinople Coast', 'Bulgaria North Coast'),
            ('Constantinople Coast', 'Ankara Coast'),
            ('Constantinople Coast', 'Smyrna Coast'),
            ('Smyrna Coast',         'Syria Coast'),
            ('Ankara Coast',         'Armenia Coast'),
            ('Armenia Coast',        'Sevastopol Coast'),
            ('Sevastopol Coast',     'Rumania Coast'),
            ('Rumania Coast',        'Bulgaria North Coast'),

            # Land Adjacencies
            ('Apulia', 'Venice'),
            ('Albania', 'Trieste'),
            ('Clyde', 'Edinburgh'),
            ('Clyde', 'Liverpool'),
            ('Liverpool', 'Edinburgh'),
            ('Liverpool', 'Yorkshire'),
            ('Liverpool', 'Wales'),
            ('Edinburgh', 'Yorkshire'),
            ('Wales', 'London'),
            ('Wales', 'Yorkshire'),
            ('Yorkshire', 'London'),
            ('Norway', 'Sweden'),
            ('Norway', 'Finland'),
            ('Norway', 'St. Petersburg'),
            ('Sweden', 'Finland'),
            ('Sweden', 'Denmark'),
            ('Finland', 'St. Petersburg'),
            ('Denmark', 'Kiel'),
            ('St. Petersburg', 'Livonia'),
            ('St. Petersburg', 'Moscow'),
            ('Livonia', 'Moscow'),
            ('Livonia', 'Warsaw'),
            ('Livonia', 'Prussia'),
            ('Moscow', 'Warsaw'),
            ('Moscow', 'Ukraine'),
            ('Moscow', 'Sevastopol'),
            ('Warsaw', 'Prussia'),
            ('Warsaw', 'Silesia'),
            ('Warsaw', 'Galicia'),
            ('Warsaw', 'Ukraine'),
            ('Ukraine', 'Galicia'),
            ('Ukraine', 'Sevastopol'),
            ('Ukraine', 'Rumania'),
            ('Sevastopol', 'Armenia'),
            ('Sevastopol', 'Rumania'),
            ('Armenia', 'Syria'),
            ('Armenia', 'Smyrna'),
            ('Armenia', 'Ankara'),
            ('Ankara', 'Smyrna'),
            ('Ankara', 'Constantinople'),
            ('Smyrna', 'Constantinople'),
            ('Smyrna', 'Syria'),
            ('Constantinople', 'Bulgaria'),
            ('Bulgaria', 'Rumania'),
            ('Bulgaria', 'Greece'),
            ('Bulgaria', 'Serbia'),
            ('Rumania', 'Galicia'),
            ('Rumania', 'Budapest'),
            ('Rumania', 'Serbia'),
            ('Greece', 'Albania'),
            ('Greece', 'Serbia'),
            ('Serbia', 'Albania'),
            ('Serbia', 'Trieste'),
            ('Serbia', 'Budapest'),
            ('Budapest', 'Trieste'),
            ('Budapest', 'Vienna'),
            ('Budapest', 'Galicia'),
            ('Galicia', 'Vienna'),
            ('Galicia', 'Bohemia'),
            ('Galicia', 'Silesia'),
            ('Silesia', 'Bohemia'),
            ('Silesia', 'Munich'),
            ('Silesia', 'Berlin'),
            ('Silesia', 'Prussia'),
            ('Prussia', 'Berlin'),
            ('Berlin', 'Kiel'),
            ('Berlin', 'Munich'),
            ('Munich', 'Kiel'),
            ('Munich', 'Ruhr'),
            ('Munich', 'Burgundy'),
            ('Munich', 'Tyrolia'),
            ('Munich', 'Bohemia'),
            ('Bohemia', 'Tyrolia'),
            ('Bohemia', 'Vienna'),
            ('Vienna', 'Tyrolia'),
            ('Vienna', 'Trieste'),
            ('Trieste', 'Venice'),
            ('Trieste', 'Tyrolia'),
            ('Kiel', 'Ruhr'),
            ('Kiel', 'Holland'),
            ('Ruhr', 'Holland'),
            ('Holland', 'Belgium'),
            ('Belgium', 'Ruhr'),
            ('Belgium', 'Burgundy'),
            ('Belgium', 'Picardy'),
            ('Burgundy', 'Ruhr'),
            ('Burgundy', 'Marseilles'),
            ('Burgundy', 'Gascony'),
            ('Burgundy', 'Paris'),
            ('Burgundy', 'Picardy'),
            ('Picardy', 'Paris'),
            ('Picardy', 'Brest'),
            ('Brest', 'Gascony'),
            ('Brest', 'Paris'),
            ('Gascony', 'Paris'),
            ('Gascony', 'Marseilles'),
            ('Gascony', 'Spain'),
            ('Spain', 'Portugal'),
            ('Spain', 'Marseilles'),
            ('North Africa', 'Tunis'),
            ('Marseilles', 'Piedmont'),
            ('Piedmont', 'Tuscany'),
            ('Piedmont', 'Venice'),
            ('Piedmont', 'Tyrolia'),
            ('Tuscany', 'Venice'),
            ('Tuscany', 'Rome'),
            ('Rome', 'Venice'),
            ('Rome', 'Apulia'),
            ('Rome', 'Naples'),
            ('Naples', 'Apulia'),
            ('Venice', 'Tyrolia'),
        ]

        _VANILLA_DIP_MAP_CACHE = Map(territory_descriptors, adjacencies)
    return deepcopy(_VANILLA_DIP_MAP_CACHE)


def generate_supply_center_map():
    global _VANILLA_DIP_SUPPLY_CENTER_MAP_CACHE
    if not _VANILLA_DIP_SUPPLY_CENTER_MAP_CACHE:
        supply_centers = {
            'Edinburgh',
            'London',
            'Liverpool',
            'Norway',
            'Sweden',
            'St. Petersburg',
            'Denmark',
            'Belgium',
            'Holland',
            'Kiel',
            'Berlin',
            'Munich',
            'Brest',
            'Paris',
            'Marseilles',
            'Portugal',
            'Spain',
            'Tunis',
            'Naples',
            'Rome',
            'Venice',
            'Trieste',
            'Vienna',
            'Budapest',
            'Serbia',
            'Greece',
            'Bulgaria',
            'Rumania',
            'Constantinople',
            'Ankara',
            'Smyrna',
            'Sevastopol',
            'Moscow',
            'Warsaw',
        }
        _VANILLA_DIP_SUPPLY_CENTER_MAP_CACHE = SupplyCenterMap(generate_map(), supply_centers)
    return deepcopy(_VANILLA_DIP_SUPPLY_CENTER_MAP_CACHE)


def generate_home_territories():
    global _VANILLA_DIP_HOME_TERRITORY_CACHE
    if not _VANILLA_DIP_HOME_TERRITORY_CACHE:
        _VANILLA_DIP_HOME_TERRITORY_CACHE = {
            'England': {
                'Liverpool',
                'London',
                'Edinburgh',
            },
            'France': {
                'Brest',
                'Paris',
                'Marseilles',
            },
            'Germany': {
                'Kiel',
                'Berlin',
                'Munich',
            },
            'Italy': {
                'Venice',
                'Rome',
                'Naples',
            },
            'Russia': {
                'St. Petersburg',
                'Moscow',
                'Warsaw',
                'Sevastopol',
            },
            'Austria': {
                'Vienna',
                'Budapest',
                'Trieste',
            },
            'Turkey': {
                'Constantinople',
                'Ankara',
                'Smyrna',
            },
        }
    return deepcopy(_VANILLA_DIP_HOME_TERRITORY_CACHE)


def generate_starting_ownership_map():
    global _VANILLA_DIP_STARTING_OWNERSHIP_MAP_CACHE
    if not _VANILLA_DIP_STARTING_OWNERSHIP_MAP_CACHE:
        # Vanilla Diplomacy uses the home territories as starting ownership
        _VANILLA_DIP_STARTING_OWNERSHIP_MAP_CACHE = OwnershipMap(
            generate_supply_center_map(),
            generate_home_territories(),
            generate_home_territories(),
        )
    return _VANILLA_DIP_STARTING_OWNERSHIP_MAP_CACHE


def generate_starting_player_units():
    global _VANILLA_DIP_STARTING_PLAYER_UNITS_CACHE
    if not _VANILLA_DIP_STARTING_PLAYER_UNITS_CACHE:
        # Vanilla Diplomacy uses the home territories as starting ownership
        _VANILLA_DIP_STARTING_PLAYER_UNITS_CACHE = {
            'England': {
                Unit(UnitTypes.TROOP, 'Liverpool'),
                Unit(UnitTypes.FLEET, 'London Coast'),
                Unit(UnitTypes.FLEET, 'Edinburgh Coast'),
            },
            'France': {
                Unit(UnitTypes.FLEET, 'Brest Coast'),
                Unit(UnitTypes.TROOP, 'Paris'),
                Unit(UnitTypes.TROOP, 'Marseilles'),
            },
            'Germany': {
                Unit(UnitTypes.FLEET, 'Kiel Coast'),
                Unit(UnitTypes.TROOP, 'Berlin'),
                Unit(UnitTypes.TROOP, 'Munich'),
            },
            'Italy': {
                Unit(UnitTypes.TROOP, 'Venice'),
                Unit(UnitTypes.TROOP, 'Rome'),
                Unit(UnitTypes.FLEET, 'Naples Coast'),
            },
            'Russia': {
                Unit(UnitTypes.FLEET, 'St. Petersburg South Coast'),
                Unit(UnitTypes.TROOP, 'Moscow'),
                Unit(UnitTypes.TROOP, 'Warsaw'),
                Unit(UnitTypes.FLEET, 'Sevastopol Coast'),
            },
            'Austria': {
                Unit(UnitTypes.TROOP, 'Vienna'),
                Unit(UnitTypes.TROOP, 'Budapest'),
                Unit(UnitTypes.FLEET, 'Trieste Coast'),
            },
            'Turkey': {
                Unit(UnitTypes.TROOP, 'Constantinople'),
                Unit(UnitTypes.FLEET, 'Ankara Coast'),
                Unit(UnitTypes.TROOP, 'Smyrna'),
            },
        }
    return deepcopy(_VANILLA_DIP_STARTING_PLAYER_UNITS_CACHE)



from pydip.map.territory import SeaTerritory
from pydip.player.helpers import (
    unit_can_enter,
    unit_can_support,
    territory_is_convoy_compatible,
)
from pydip.player.unit import UnitTypes


class Command:
    """ Player -- who is issuing command """
    player = None

    """ Unit -- unit being issued command """
    unit = None

    def __init__(self, player, unit):
        assert unit in player.units
        self.player = player
        self.unit = unit

    def __eq__(self, other):
        return ((isinstance(other, Command)) and
                (self.player.name == other.player.name) and
                (self.unit == other.unit))

    def __ne__(self, other):
        return not self.__eq__(other)


class MoveCommand(Command):
    """
    String -- Name of territory being moved to.
      * Must be adjacent or equal to unit's current territory
      * Must be legal for unit to enter this territory
    """
    destination = None

    def __init__(self, player, unit, destination):
        super().__init__(player, unit)
        game_map = self.player.game_map
        assert destination in game_map.name_map
        assert (
            unit_can_enter(game_map, unit, game_map.name_map[destination]) or
            unit.position == destination
        )
        self.destination = destination

    def __eq__(self, other):
        return (super(MoveCommand, self).__eq__(other) and
                isinstance(other, MoveCommand) and
                self.destination == other.destination)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{}: {} {} -> {}'.format(
            self.player.name,
            self.unit.unit_type,
            self.unit.position,
            self.destination,
        )


class HoldCommand(MoveCommand):
    """ Holding is really just moving to your current position """
    def __init__(self, player, unit):
        super().__init__(player, unit, unit.position)

    def __repr__(self):
        return '{}: {} {} Hold'.format(
            self.player.name,
            self.unit.unit_type,
            self.unit.position,
        )


class SupportCommand(Command):
    """ Unit -- unit to support """
    supported_unit = None

    """
    String -- name of territory to support into.
      * Must be adjacent to unit's current territory
      * Must be legal for unit to enter this territory
      * Must be adjacent or identical to supported_unit's current territory
      * Must be legal for supported_unit to support this territory
    """
    destination = None

    def __init__(self, player, unit, supported_unit, destination):
        super().__init__(player, unit)
        self.supported_unit = supported_unit
        self.destination = destination
        game_map = self.player.game_map

        assert destination in game_map.name_map
        destination_territory = game_map.name_map[destination]
        supported_territory   = game_map.name_map[supported_unit.position]

        assert (
            destination == supported_unit.position or
            unit_can_enter(game_map, supported_unit, destination_territory) or
            (supported_unit.unit_type == UnitTypes.TROOP and
             territory_is_convoy_compatible(destination_territory) and
             territory_is_convoy_compatible(supported_territory))
        )
        assert unit_can_support(game_map, unit, destination_territory)

    def __eq__(self, other):
        return (super(SupportCommand, self).__eq__(other) and
                isinstance(other, SupportCommand) and
                self.supported_unit == other.supported_unit and
                self.destination == other.desination)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{}: {} {} Supports {} {} -> {}'.format(
            self.player.name,
            self.unit.unit_type,
            self.unit.position,
            self.supported_unit.unit_type,
            self.supported_unit.position,
            self.destination,
        )


class ConvoyMoveCommand(Command):
    """
    String -- name of territory convoying to.
      * Must be LandTerritory with at least one coast
    """
    destination = None

    """
    Extra Assertions:
      * unit must be a TROOP
      * unit's current territory must be LandTerritory with at least one coast
      * convoys with destinations equal to the source are illegal
    """

    def __init__(self, player, unit, destination):
        super().__init__(player, unit)
        self.destination = destination
        game_map = self.player.game_map

        current_territory     = game_map.name_map[unit.position]
        destination_territory = game_map.name_map[destination]
        assert current_territory != destination_territory

        assert unit.unit_type == UnitTypes.TROOP
        assert territory_is_convoy_compatible(current_territory)

        assert destination in game_map.name_map
        assert territory_is_convoy_compatible(destination_territory)

    def __eq__(self, other):
        return (super(ConvoyMoveCommand, self).__eq__(other) and
                isinstance(other, ConvoyMoveCommand) and
                self.destination == other.destination)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{}: {} {} -> {} (Convoy)'.format(
            self.player.name,
            self.unit.unit_type,
            self.unit.position,
            self.destination,
        )


class ConvoyTransportCommand(Command):
    """
    Unit -- unit to convoy.
      * Must be a TROOP
    """
    transported_unit = None

    """
    String -- name of territory convoying to.
      * Must be LandTerritory with at least one coast
    """
    destination = None

    """
    Extra Assertions:
      * unit must be a FLEET
      * unit's current territory must be a SeaTerritory
      * transported unit's current territory must be LandTerritory with at least one coast
      * convoys with destinations equal to the source are illegal
    """

    def __init__(self, player, unit, transported_unit, destination):
        super().__init__(player, unit)
        self.transported_unit = transported_unit
        self.destination = destination
        assert unit.unit_type == UnitTypes.FLEET
        game_map = self.player.game_map

        current_territory = game_map.name_map[unit.position]
        source_territory = game_map.name_map[transported_unit.position]
        destination_territory = game_map.name_map[destination]
        assert source_territory != destination_territory
        assert isinstance(current_territory, SeaTerritory)

        assert transported_unit.unit_type == UnitTypes.TROOP
        assert territory_is_convoy_compatible(source_territory)

        assert destination in game_map.name_map
        assert territory_is_convoy_compatible(destination_territory)

    def __eq__(self, other):
        return (super(ConvoyTransportCommand, self).__eq__(other) and
                isinstance(other, ConvoyTransportCommand) and
                self.transported_unit == other.transported_unit and
                self.destination == other.desination)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{}: {} {} Transport {} {} -> {} (Convoy)'.format(
            self.player.name,
            self.unit.unit_type,
            self.unit.position,
            self.transported_unit.unit_type,
            self.transported_unit.position,
            self.destination,
        )



from pydip.player.helpers import unit_type_can_enter


class AdjustmentCommand:
    def __init__(self, player, unit):
        self.player = player
        self.unit = unit


class AdjustmentDisbandCommand(AdjustmentCommand):
    def __init__(self, player, unit):
        super().__init__(player, unit)
        assert unit in player.units

    def __repr__(self):
        return '{}: {} {} Disband'.format(
            self.player.name,
            self.unit.unit_type,
            self.unit.position,
        )

    def __eq__(self, other):
        return (super(AdjustmentDisbandCommand, self).__eq__(other) and
                isinstance(other, AdjustmentDisbandCommand))

    def __ne__(self, other):
        return not self.__eq__(other)


class AdjustmentCreateCommand(AdjustmentCommand):
    """
    Unlike a traditional command, this one uses the unit to represent the new unit to be created.
    So this command takes an OwnershipMap, representing current board state, to determine if the
    provided unit would be allowed to be created for this player.
    """
    def __init__(self, ownership_map, player, unit):
        super().__init__(player, unit)

        unit_territory = ownership_map.supply_map.game_map.name_map[unit.position]
        assert ownership_map.territory_is_owned(player.name, unit.position)
        assert ownership_map.territory_is_home(player.name, unit.position)
        assert unit_type_can_enter(unit.unit_type, unit_territory)

        for existing_unit in player.units:
            existing_unit_territory = ownership_map.supply_map.game_map.name_map[existing_unit.position]
            assert not unit_territory.same_territory(existing_unit_territory)

    def __eq__(self, other):
        return (super(AdjustmentCreateCommand, self).__eq__(other) and
                isinstance(other, AdjustmentCreateCommand))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{}: {} {} Create'.format(
            self.player.name,
            self.unit.unit_type,
            self.unit.position,
        )



from pydip.map.territory import CoastTerritory, SeaTerritory, LandTerritory
from pydip.player.unit import UnitTypes


def unit_can_enter(game_map, unit, territory):
    if territory.name not in game_map.adjacency[unit.position]:
        return False
    else:
        return unit_type_can_enter(unit.unit_type, territory)


def unit_type_can_enter(unit_type, territory):
    if unit_type == UnitTypes.TROOP:
        return isinstance(territory, LandTerritory)
    elif unit_type == UnitTypes.FLEET:
        return isinstance(territory, SeaTerritory) or isinstance(territory, CoastTerritory)
    else:
        raise ValueError("Invalid UnitType: {}".format(unit_type))


def unit_can_support(game_map, unit, territory):
    """
    Being able to enter a coast or a land territory allows you to support,
    so we'll check each possible option
    """
    territories_to_check = [ territory ]
    if isinstance(territory, LandTerritory):
        territories_to_check.extend(territory.coasts)
    elif isinstance(territory, CoastTerritory):
        territories_to_check.append(territory.parent)
        territories_to_check.extend(territory.parent.coasts)

    return any(unit_can_enter(game_map, unit, to_check) for to_check in territories_to_check)


def territory_is_convoy_compatible(territory):
    return isinstance(territory, LandTerritory) and len(territory.coasts) > 0



from copy import deepcopy

from pydip.map.map import OwnershipMap
from pydip.player.command.adjustment_command import AdjustmentCommand, AdjustmentCreateCommand, AdjustmentDisbandCommand


def calculate_adjustments(ownership_map, player_units):
    """
    Given a map of current ownership, and a map of new unit positions (after
    retreats have been resolved), this function will compute a new ownership
    map as well as what adjustments each player must perform. NOTE: this
    function has no knowledge of Spring vs Fall turns, so the caller should
    only use this after retreats in the Fall.

    ownership_map should be an OwnershipMap for the current game.

    player_units should be the output from turn.retreat.resolve_retreats --
    that is, a map of player names to sets of units after retreat resolution.

    This function returns two values: First, a new OwnershipMap updated for
    new unit positions. Second, a map of player names to integers which
    represent their adjustment requirements. Positive integers denote players
    who are allowed to build units this turn. Negative integers indicate the
    player must disband units this turn.
    """
    all_players = ownership_map.owned_territories.keys()
    new_owned_territories = { player : set() for player in all_players }

    for player in all_players:
        other_player_positions = set()
        for other_player in all_players:
            if other_player != player:
                other_player_positions |= {
                    ownership_map.supply_map.game_map.relevant_name_for_territory(unit.position)
                    for unit in player_units[other_player]
                }

        for territory in ownership_map.owned_territories[player]:
            if territory not in other_player_positions:
                if territory in ownership_map.supply_map.supply_centers:
                    new_owned_territories[player].add(territory)

        for unit in player_units[player]:
            relevant_territory = ownership_map.supply_map.game_map.relevant_name_for_territory(unit.position)
            if relevant_territory in ownership_map.supply_map.supply_centers:
                new_owned_territories[player].add(relevant_territory)

    new_ownership_map = OwnershipMap(
        ownership_map.supply_map,
        new_owned_territories,
        ownership_map.home_territories)

    adjustment_counts = {
        player : len(new_ownership_map.owned_territories[player]) - len(player_units[player])
        for player in all_players
    }

    return new_ownership_map, adjustment_counts


def resolve_adjustment__validated(ownership_map, adjustment_counts, player_units, commands):
    """
    Given an OwnershipMap representing the current board state, a mapping of adjustment expectations
    (expected to be output from calcluate_adjustments), a collection of current player units, and a
    list of AdjustmentCommands to issue, resolves the adjustments and determines what units each
    player will end up with at the end of the turn.

    This specific function also validates several properties that violate strict adherence to Civil
    Disobedience rules. This system typically asserts that users are not allowed to issue illegal
    command sets, and that includes not issuing commands that are required (e.g. Disbanding during
    unit adjustment). However, forcing the caller to follow these atypical policies can potentially
    lead to confusion and invalid game states.

    To that end, there are two adjustment functions provided: one with strict validation, forcing the
    caller to perform sanity checks (and potentially implement Civil Disobedience policies) themselves,
    and a less strictly validated version that applies Civil Disobedience policies automatically when
    improper order counts are provided.


    Both versions return a mapping of player names to sets of Units, representing the units owned by
    each player after adjustments have been resolved.
    """
    _validate_adjustments(ownership_map, adjustment_counts, commands)
    resolve_adjustment(ownership_map, adjustment_counts, player_units, commands)


def resolve_adjustment(ownership_map, adjustment_counts, player_units, commands):
    new_player_units = deepcopy(player_units)
    for player, count in adjustment_counts.items():
        if count < 0:
            disbands = _find_disbands(player, adjustment_counts, player_units, commands)
            new_player_units[player] -= disbands
        elif count > 0:
            creates = _find_creates(ownership_map.supply_map.game_map, player, adjustment_counts, commands)
            new_player_units[player] |= creates

    return new_player_units


def _find_disbands(player, adjustment_counts, player_units, commands):
    expected_disband_count = -adjustment_counts[player]

    filtered_commands = filter(lambda c: isinstance(c, AdjustmentDisbandCommand), commands)
    filtered_commands = filter(lambda c: c.player.name == player, filtered_commands)
    filtered_commands = list(filtered_commands)

    # if too many disbands, truncate
    if len(filtered_commands) > expected_disband_count:
        filtered_commands = filtered_commands[:expected_disband_count]
    disbands = { command.unit for command in filtered_commands }

    # if too few disbands, begin disbanding units in alphabetical order
    while len(disbands) < expected_disband_count:
        remaining_units = sorted(
            list(set(player_units[player]) - disbands),
            key = lambda u: u.position.lower()
        )
        disbands.add(remaining_units[0])

    return disbands


def _find_creates(game_map, player, adjustment_counts, commands):
    permitted_create_count = adjustment_counts[player]

    filtered_commands = filter(lambda c: isinstance(c, AdjustmentCreateCommand), commands)
    filtered_commands = filter(lambda c: c.player.name == player, filtered_commands)
    filtered_commands = list(filtered_commands)

    # if two creates on same territory, keep first only
    territories_with_creation = set()
    preserved_commands = []
    for command in filtered_commands:
        relevant_name = game_map.relevant_name_for_territory(command.unit.position)
        if relevant_name in territories_with_creation:
            continue
        territories_with_creation.add(relevant_name)
        preserved_commands.append(command)

    # if too many creates, truncate
    if len(preserved_commands) > permitted_create_count:
        preserved_commands = preserved_commands[:permitted_create_count]
    return { command.unit for command in preserved_commands }


def _validate_adjustments(ownership_map, adjustment_counts, commands):
    assert all(isinstance(command, AdjustmentCommand) for command in commands)

    provided_adjustment_terriories = dict()
    for command in commands:
        player = command.player.name
        game_map = ownership_map.supply_map.game_map
        if isinstance(command, AdjustmentCreateCommand):
            if player not in provided_adjustment_terriories:
                provided_adjustment_terriories[player] = set()
            elif adjustment_counts[player] < 0:
                raise AssertionError("{} attempting to Create when required to Disband".format(player))
            elif len(provided_adjustment_terriories[player]) >= adjustment_counts[player]:
                raise AssertionError("{} attempting to Create too many units in one turn".format(player))

            relevant_command_territory = game_map.relevant_name_for_territory(command.unit.position)
            if relevant_command_territory in provided_adjustment_terriories[player]:
                raise AssertionError("{} attempting to Create multiple units in same territory".format(player))

            provided_adjustment_terriories[player].add(relevant_command_territory)
        elif isinstance(command, AdjustmentDisbandCommand):
            if player not in provided_adjustment_terriories:
                provided_adjustment_terriories[player] = set()
            elif adjustment_counts[player] > 0:
                raise AssertionError("{} attempting to Disband when expected to Create".format(player))
            elif len(provided_adjustment_terriories[player]) >= -adjustment_counts[player]:
                raise AssertionError("{} attempting to Disband too many units in one turn".format(player))

            relevant_command_territory = game_map.relevant_name_for_territory(command.unit.position)
            if relevant_command_territory in provided_adjustment_terriories[player]:
                raise AssertionError("{} attempting to Disband same unit multiple times".format(player))

            provided_adjustment_terriories[player].add(relevant_command_territory)
        else:
            raise AssertionError("Unexpected AdjustmentCommand type")

    for player, count in adjustment_counts.items():
        if count < 0:
            if len(provided_adjustment_terriories.get(player, set())) != -count:
                raise AssertionError("{} failed to Disband sufficient units this turn".format(player))



from distutils.core import setup
from setuptools import find_packages
setup(
    name = 'pydip',
    packages = find_packages(exclude=['pydip.test', 'pydip.test.*']),
    version = '0.1.7',
    description = 'Adjudication logic engine for Diplomacy board game',
    author = 'Aric Parkinson',
    author_email = 'aric.parkinson@gmail.com',
    url = 'https://github.com/aparkins/pydip',
    download_url = 'https://github.com/aparkins/pydip/tarball/0.1.6',
    keywords = ['diplomacy', 'adjudication', 'adjudicator', 'board', 'game'],
    classifiers = [],
    tests_require = ['pytest'],
)



from pydip.player.command.command import Command


class RetreatCommand(Command):
    """
    retreat_map is intended to be the output from turn.resolve.resolve_turn.
    That is, a mapping of players to mappings of units to either None (in the
    case that a retreat is not expected for that unit), or a set of territory
    names (which hold the valid retreat targets).

    Commands issued that disagree with the provided retreat_map will fail.
    """
    def __init__(self, retreat_map, player, unit):
        super().__init__(player, unit)
        assert unit in retreat_map[player.name]
        assert retreat_map[player.name][unit] is not None


class RetreatDisbandCommand(RetreatCommand):
    def __init__(self, retreat_map, player, unit):
        super().__init__(retreat_map, player, unit)

    def __repr__(self):
        return '{}: {} {} Disband'.format(
            self.player.name,
            self.unit.unit_type,
            self.unit.position,
        )

    def __eq__(self, other):
        return (super(RetreatDisbandCommand, self).__eq__(other) and
                isinstance(other, RetreatDisbandCommand))

    def __ne__(self, other):
        return not self.__eq__(other)


class RetreatMoveCommand(RetreatCommand):
    """
    String -- Name of territory being retreated to.
      * Must be included in retreat_map as one of the territories the unit
        is permitted to retreat to
    """
    destination = None

    def __init__(self, retreat_map, player, unit, destination):
        super().__init__(retreat_map, player, unit)
        game_map = self.player.game_map
        assert destination in game_map.name_map
        assert destination in retreat_map[player.name][unit]
        self.destination = destination

    def __eq__(self, other):
        return (super(RetreatMoveCommand, self).__eq__(other) and
                isinstance(other, RetreatMoveCommand) and
                self.destination == other.destination)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{}: {} {} -> {} (Retreat)'.format(
            self.player.name,
            self.unit.unit_type,
            self.unit.position,
            self.destination,
        )



from collections import defaultdict
from enum import Enum
from functools import reduce

from pydip.map.territory import CoastTerritory, SeaTerritory, LandTerritory
from pydip.player.command.command import MoveCommand, ConvoyMoveCommand, ConvoyTransportCommand, SupportCommand
from pydip.player.unit import Unit
from pydip.turn.command_map import CommandMap


def resolve_turn(game_map, commands):
    """
    Returns resulting positions of each unit by considering interactions
    of provided list of commands.

    Note: assumes that every unit on the map has been issued a command.
    If a unit has _not_ been issued a command, you ought to insert a
    HoldCommand for that unit by default.

    map: Map representing game board
    commands: Command[] representing each command issued for this turn
              Expected to be fully populated for each unit on the board
              (Recommended to default to Hold for units not given orders)

    returns a map of Player names to sub-maps representing results of that
    player's units. The sub-maps will be maps of Units to an optional set
    of territories, representing which territories the unit can retreat to.
    If the territory set is not provided, no retreat is required. If the
    territory set is provided, a retreat is required. If it is empty, no
    retreat is possible.
    """
    command_map = CommandMap(game_map, commands)
    _init_resolution()
    resolutions = {command.unit.position: _resolve(game_map, command_map, command) for command in commands}
    return compute_retreats(game_map, command_map, commands, resolutions)


#------------------------------------------------------------------------------
# Below implementation is taken with minor modification from Lucas Kruijswijk:
# http://www.diplom.org/Zine/S2009M/Kruijswijk/DipMath_Chp6.htm
#------------------------------------------------------------------------------


class ResolutionState(Enum):
    UNRESOLVED = 0
    GUESSING   = 1
    RESOLVED   = 2

resolution_map  = None
state_map       = None
dependency_list = None


def _init_resolution():
    global resolution_map
    global state_map
    global dependency_list
    resolution_map = defaultdict(bool)
    state_map      = defaultdict(lambda: ResolutionState.UNRESOLVED)
    dependency_list = list()


def _resolve(game_map, command_map, command):
    global resolution_map
    global state_map
    global dependency_list
    command_territory = command.unit.position

    if state_map[command_territory] == ResolutionState.RESOLVED:
        return resolution_map[command_territory]

    if state_map[command_territory] == ResolutionState.GUESSING:
        if command_territory not in dependency_list:
            dependency_list.append(command_territory)
        return resolution_map[command_territory]

    old_dependency_length = len(dependency_list)

    # Initially, guess that we fail
    resolution_map[command_territory] = False
    state_map[command_territory]      = ResolutionState.GUESSING
    fail_guess_result                 = _adjudicate(game_map, command_map, command)

    # If the dependency graph didn't change as a consequence, our result doesn't
    # depend on the guess and we can return right away
    if old_dependency_length == len(dependency_list):
        # This is possible because of the backup rule in paradox resolutions
        if state_map[command_territory] != ResolutionState.RESOLVED:
            resolution_map[command_territory] = fail_guess_result
            state_map[command_territory]      = ResolutionState.RESOLVED

        return fail_guess_result

    dependency_sub_set = set(dependency_list[old_dependency_length:])
    # If we don't depend on ourselves yet, we add ourselves in to complete the
    # cycle, and let our caller sort out the details
    if command_territory not in dependency_sub_set:
        dependency_list.append(command_territory)
        resolution_map[command_territory] = fail_guess_result
        return fail_guess_result

    # Otherwise, we depend on our own guess, so we need to clear out dependencies
    # to check the other guess for consistency
    for dependent_territory in dependency_sub_set:
        state_map[dependent_territory] = ResolutionState.UNRESOLVED
    dependency_list = dependency_list[:old_dependency_length]

    resolution_map[command_territory] = True
    state_map[command_territory]      = ResolutionState.GUESSING
    success_guess_result              = _adjudicate(game_map, command_map, command)

    # If results are consistent, no need for further checking
    if fail_guess_result == success_guess_result:
        for dependent_territory in dependency_sub_set:
            state_map[dependent_territory] = ResolutionState.UNRESOLVED
        dependency_list = dependency_list[:old_dependency_length]

        resolution_map[command_territory] = fail_guess_result
        state_map[command_territory]      = ResolutionState.RESOLVED
        return fail_guess_result

    # If we got to this point, that means we encountered a paradox that has two
    # consistent outcomes, and we need a backup rule to fully resolve it
    _backup_rule(command_map, dependency_sub_set)
    dependency_list = dependency_list[:old_dependency_length]

    # And because the backup rule may not resolve our own command, we'll need to
    # start fresh just to be sure
    return _resolve(game_map, command_map, command)


def _adjudicate(game_map, command_map, command):
    if isinstance(command, MoveCommand):
        return _adjudicate_move(game_map, command_map, command)
    elif isinstance(command, ConvoyMoveCommand):
        return _adjudicate_convoy_move(game_map, command_map, command)
    elif isinstance(command, ConvoyTransportCommand):
        return _adjudicate_convoy_transport(game_map, command_map, command)
    elif isinstance(command, SupportCommand):
        return _adjudicate_support(game_map, command_map, command)
    else:
        raise ValueError("Command unexpected type")


def _backup_rule(command_map, dependency_set):
    global resolution_map
    global state_map
    for dependency_territory in dependency_set:
        dependency = command_map.get_home_command(dependency_territory)
        if isinstance(dependency, MoveCommand):
            if isinstance(command_map.get_home_command(dependency.destination), ConvoyTransportCommand):
                _apply_szykman(command_map, dependency_set)
                return
    _apply_circular_movement(command_map, dependency_set)


def _apply_szykman(command_map, dependency_set):
    for dependency_territory in dependency_set:
        dependency = command_map.get_home_command(dependency_territory)
        if isinstance(dependency, ConvoyMoveCommand) or isinstance(dependency, ConvoyTransportCommand):
            resolution_map[dependency_territory] = False
            state_map[dependency_territory] = ResolutionState.RESOLVED
        else:
            state_map[dependency_territory] = ResolutionState.UNRESOLVED


def _apply_circular_movement(command_map, dependency_set):
    for dependency_territory in dependency_set:
        dependency = command_map.get_home_command(dependency_territory)
        if isinstance(dependency, MoveCommand) or isinstance(dependency, ConvoyMoveCommand):
            resolution_map[dependency_territory] = True
            state_map[dependency_territory] = ResolutionState.RESOLVED
        else:
            state_map[dependency_territory] = ResolutionState.UNRESOLVED


#----------------------
# convoys
#----------------------
def _adjudicate_convoy_move(game_map, command_map, command):
    assert isinstance(command, ConvoyMoveCommand)
    if not _has_path(game_map, command_map, command):
        return False
    return _adjudicate_move(game_map, command_map, command)


def _adjudicate_convoy_transport(game_map, command_map, command):
    assert isinstance(command, ConvoyTransportCommand)
    return not _is_dislodged(game_map, command_map, command.unit)


def _has_path(game_map, command_map, command):
    assert isinstance(command, ConvoyMoveCommand)
    visited     = set()
    source      = command.unit.position
    destination = command.destination
    possible_transports = command_map.get_convoy_transports(source, destination)
    possible_transport_territories = { transport.unit.position for transport in possible_transports }

    starting_territory = game_map.name_map[source]
    coastal_adjacencies = reduce(set.__or__, [game_map.adjacency[coast.name] for coast in starting_territory.coasts])
    coastal_adjacencies = list(filter(lambda t: isinstance(game_map.name_map[t], SeaTerritory), coastal_adjacencies))
    to_visit = [
        possible_transport_territory
        for possible_transport_territory in possible_transport_territories
        if any(possible_transport_territory in coastal_adjacency for coastal_adjacency in coastal_adjacencies)
    ]

    while len(to_visit) > 0:
        visiting = to_visit.pop()
        visited.add(visiting)
        # if the convoy was disrupted, we can't use it as part of our chain
        if not _resolve(game_map, command_map, command_map.get_home_command(visiting)):
            continue

        adjacent = [game_map.name_map[adj] for adj in game_map.adjacency[visiting]]

        adjacent_land = {
            coast.parent.name for coast in adjacent
            if isinstance(coast, CoastTerritory)
        }
        if destination in adjacent_land:
            return True

        to_visit = [
           territory.name for territory in adjacent
           if (territory.name not in visited and
               territory.name in possible_transport_territories)
        ] + to_visit

    return False


#----------------------
# move
#----------------------
def _adjudicate_move(game_map, command_map, command):
    assert isinstance(command, MoveCommand) or isinstance(command, ConvoyMoveCommand)

    attack_strength = _attack_strength(game_map, command_map, command)

    prevent_combatants     = _get_prevent_combatants(command_map, command)
    high_prevent_strength = max([_prevent_strength(game_map, command_map, prevent_combatant)
                                 for prevent_combatant in prevent_combatants] + [0])
    if attack_strength <= high_prevent_strength:
        return False

    head_to_head_combatant = _get_head_to_head_combatant(game_map, command_map, command)
    if head_to_head_combatant is not None:
        return attack_strength > _defend_strength(game_map, command_map, head_to_head_combatant)
    return attack_strength > _hold_strength(game_map, command_map, command.destination)


def _get_prevent_combatants(command_map, command):
    direct_combatants = command_map.get_attackers(command.destination)
    convoy_combatants = command_map.get_convoy_attackers(command.destination)
    combatants = direct_combatants + convoy_combatants
    combatants = filter(lambda c: c != command, combatants)
    return combatants


def _get_head_to_head_combatant(game_map, command_map, command):
    if isinstance(command, ConvoyMoveCommand):
        return None
    potential_attacker = command_map.get_home_command(command.destination)
    if potential_attacker is not None:
        if isinstance(potential_attacker, MoveCommand):
            attacked_territory = game_map.name_map[potential_attacker.destination]
            position_territory = game_map.name_map[command.unit.position]
            if position_territory.same_territory(attacked_territory):
                return potential_attacker
    return None


def _attack_strength(game_map, command_map, command):
    if isinstance(command, ConvoyMoveCommand):
        if not _has_path(game_map, command_map, command):
            return 0
    attacked_command = command_map.get_home_command(command.destination)
    supporters = command_map.get_supports(command.unit.position, command.destination)
    supporters = filter(lambda c: _resolve(game_map, command_map, c), supporters)

    if attacked_command is None:
        return 1 + len(list(supporters))
    if _get_head_to_head_combatant(game_map, command_map, command) is None:
        if isinstance(attacked_command, MoveCommand) or isinstance(attacked_command, ConvoyMoveCommand):
            if _resolve(game_map, command_map, attacked_command):
                return 1 + len(list(supporters))
    if attacked_command.player.name == command.player.name:
        return 0

    supporters = filter(lambda c: c.player.name != attacked_command.player.name, supporters)
    return 1 + len(list(supporters))


def _prevent_strength(game_map, command_map, command):
    if isinstance(command, ConvoyMoveCommand):
        if not _has_path(game_map, command_map, command):
            return 0
    head_to_head_combatant = _get_head_to_head_combatant(game_map, command_map, command)
    if head_to_head_combatant is not None:
        if _resolve(game_map, command_map, head_to_head_combatant):
            return 0

    supporters = command_map.get_supports(command.unit.position, command.destination)
    supporters = filter(lambda c: _resolve(game_map, command_map, c), supporters)
    return 1 + len(list(supporters))


def _defend_strength(game_map, command_map, command):
    supporters = command_map.get_supports(command.unit.position, command.destination)
    supporters = filter(lambda c: _resolve(game_map, command_map, c), supporters)
    return 1 + len(list(supporters))


def _hold_strength(game_map, command_map, territory):
    home_command = command_map.get_home_command(territory)
    if home_command is None:
        return 0
    if isinstance(home_command, MoveCommand) or isinstance(home_command, ConvoyMoveCommand):
        return 0 if _resolve(game_map, command_map, home_command) else 1
    supporters = command_map.get_supports(territory, territory)
    supporters = filter(lambda c: _resolve(game_map, command_map, c), supporters)
    return 1 + len(list(supporters))


#----------------------
# support
#----------------------
def _adjudicate_support(game_map, command_map, command):
    assert isinstance(command, SupportCommand)
    if _invalid_support(command_map, command):
        return False
    if len(_indirect_non_convoy_attackers(command_map, command)) > 0:
        return False
    for convoy_attacker in _indirect_convoy_attackers(command_map, command):
        if _has_path(game_map, command_map, convoy_attacker):
            return False
    return not _is_dislodged(game_map, command_map, command.unit)


def _invalid_support(command_map, command):
    supported_command = command_map.get_home_command(command.supported_unit.position)
    if isinstance(supported_command, MoveCommand) or isinstance(supported_command, ConvoyMoveCommand):
        return supported_command.destination != command.destination
    return command.destination != command.supported_unit.position


def _indirect_non_convoy_attackers(command_map, command):
    filtered = command_map.get_attackers(command.unit.position)
    filtered = filter(lambda c: c.unit.position != command.destination, filtered)
    filtered = filter(lambda c: c.player != command.player, filtered)
    return list(filtered)


def _indirect_convoy_attackers(command_map, command):
    filtered = command_map.get_convoy_attackers(command.unit.position)
    filtered = filter(lambda c: c.unit.position != command.destination, filtered)
    filtered = filter(lambda c: c.player.name != command.player.name, filtered)
    return list(filtered)


def _is_dislodged(game_map, command_map, unit):
    """
    Determines if the unit will be dislodged by a different move, assuming it stays in place.
    Please note that this function does not indicate whether the unit _will_ stay in place.
    """
    return any((_resolve(game_map, command_map, attack) for attack in _attackers(game_map, command_map, unit)))


def _attackers(game_map, command_map, unit):
    filtered = command_map.get_attackers(unit.position)
    filtered = filter(
        lambda c: isinstance(c, MoveCommand) or isinstance(c, ConvoyMoveCommand),
        filtered,
    )
    filtered = filter(
        lambda c: game_map.name_map[c.destination].same_territory(game_map.name_map[unit.position]),
        filtered,
    )

    return list(filtered)


#----------------------
# Retreats
#----------------------
def compute_retreats(game_map, command_map, commands, resolutions):
    player_results = defaultdict(dict)
    occupied_territories = _get_occupations(game_map, commands, resolutions)

    for command in commands:
        current_position = command.unit.position
        if resolutions[current_position]:
            if isinstance(command, MoveCommand) or isinstance(command, ConvoyMoveCommand):
                moved_unit = Unit(command.unit.unit_type, command.destination)
                player_results[command.player.name][moved_unit] = None
            else:
                player_results[command.player.name][command.unit] = None
        else:
            direct_attackers = list(filter(
                lambda c: resolutions[c.unit.position],
                command_map.get_attackers(current_position),
            ))
            convoy_attackers = list(filter(
                lambda c: resolutions[c.unit.position],
                command_map.get_convoy_attackers(current_position),
            ))
            attackers = direct_attackers + convoy_attackers
            if len(attackers) == 0:
                player_results[command.player.name][command.unit] = None
            else:
                retreat_options = game_map.adjacency[current_position]
                retreat_options = filter(
                    lambda t: t not in occupied_territories,
                    retreat_options,
                )
                retreat_options = filter(
                    lambda t: all(t not in _applicable_territories(game_map, attacker.unit.position)
                                  for attacker in direct_attackers),
                    retreat_options,
                )
                retreat_options = filter(
                    lambda t: _hold_strength(game_map, command_map, t) == 0,
                    retreat_options,
                )
                retreat_options = filter(
                    lambda t: all(_prevent_strength(game_map, command_map, attacker) == 0
                                  for attacker in command_map.get_attackers(t)),
                    retreat_options,
                )

                player_results[command.player.name][command.unit] = set(retreat_options)

    return player_results


def _applicable_territories(game_map, territory_name):
    territory = game_map.name_map[territory_name]
    applicable = { territory_name }
    if isinstance(territory, LandTerritory):
        applicable |= { coast.name for coast in territory.coasts }
    elif isinstance(territory, CoastTerritory):
        applicable.add(territory.parent.name)
        applicable |= { coast.name for coast in territory.parent.coasts }

    return applicable


def _get_occupations(game_map, commands, resolutions):
    occupations = set()
    for command in commands:
        if isinstance(command, MoveCommand) or isinstance(command, ConvoyMoveCommand):
            if resolutions[command.unit.position]:
                occupations |= _applicable_territories(game_map, command.destination)
            else:
                occupations |= _applicable_territories(game_map, command.unit.position)
        else:
            occupations |= _applicable_territories(game_map, command.unit.position)

    return occupations


def _attacked_by_other_unit(command_map, command, territory):
    attackers = command_map.get_attackers(territory) + command_map.get_convoy_attackers(territory)
    attackers = filter(lambda c: c != command, attackers)
    return len(list(attackers)) > 0
